﻿using System.Reflection;
using System.Windows.Forms;
using System;
using System.IO;
using LazyIO;
using System.Collections.Generic;
using System.Collections;

namespace HaloMap
{
    public abstract class Chunk : LazyIO.LazyIO, IMagicReadWrite
    {
        #region -  properties   -

        public int Magic
        {
            get;
            set;
        }
        public int Size
        {
            get
            {
                return GetSize();
            }
        }

        public List<IMetaItem> ListOfItems 
        { 
            get { return Items; } 
        }

        #endregion

        #region -    methods    -

        int GetSize()
        {
            int size = 0;
            foreach( IMetaItem item in Items )
            {
                size += item.Size;
            }
            return size;
        }
        public sealed override void Read(EndianReader br)
        {
            foreach( IMetaItem item in Items )
            {
                Type t = item.GetType( );
                if( t.IsGenericType && t.GetGenericTypeDefinition( ) == typeof(Reflexive<>) )
                {
                    IMagicReadWrite reflexive = (IMagicReadWrite)item;
                    reflexive.Read(br, Magic);
                }
                else
                {
                    item.Read(br);
                }
            }
        }
        public void Read(EndianReader br, int magic)
        {
            Magic = magic;
            Read(br);
        }
        public void PopulatePanel(HaloPanel panel, ViewType viewing, TagList tags, StringList strings)
        {
            ViewType invisibles = ViewType.Invisibles & viewing;
            bool showInvisibles = invisibles == ViewType.Invisibles;
            panel.BeginUpdate( );
            panel.Visible = false;
            panel.Controls.Clear( );
            foreach( IMetaItem mi in Items )
            {
                // check if it's supposed to be visible or we're showing invisibles
                bool visible = mi.Visible || showInvisibles;
                if( visible == false ) { continue; }

                if( mi is LazyIO.Byte )
                {
                    ViewType type = ViewType.Byte & viewing;
                    bool showThisType = type == ViewType.Byte;
                    if( showThisType == false ) // if we don't want to view it
                        continue; // go to the next meta item
                    NumberViewer<byte> viewer = new NumberViewer<byte>( );
                    viewer.SetObject((LazyIO.Byte)mi);
                    panel.Controls.Add(viewer);
                    viewer.BringToFront( );
                }
                else if( mi is LazyIO.Short )
                {
                    ViewType type = ViewType.Short & viewing;
                    bool showThisType = type == ViewType.Short;
                    if( showThisType == false )
                        continue;
                    NumberViewer<short> viewer = new NumberViewer<short>( );
                    viewer.SetObject((LazyIO.Short)mi);
                    panel.Controls.Add(viewer);
                    viewer.BringToFront( );
                }
                else if( mi is LazyIO.UShort )
                {
                    ViewType type = ViewType.UShort & viewing;
                    bool showThisType = type == ViewType.UShort;
                    if( showThisType == false )
                        continue;
                    NumberViewer<ushort> viewer = new NumberViewer<ushort>( );
                    viewer.SetObject((LazyIO.UShort)mi);
                    panel.Controls.Add(viewer);
                    viewer.BringToFront( );
                }
                else if( mi is LazyIO.Int )
                {
                    ViewType type = ViewType.Int & viewing;
                    bool showThisType = type == ViewType.Int;
                    if( showThisType == false )
                        continue;
                    NumberViewer<int> viewer = new NumberViewer<int>( );
                    viewer.SetObject((LazyIO.Int)mi);
                    panel.Controls.Add(viewer);
                    viewer.BringToFront( );
                }
                else if( mi is LazyIO.UInt )
                {
                    ViewType type = ViewType.UInt & viewing;
                    bool showThisType = type == ViewType.UInt;
                    if( showThisType == false )
                        continue;
                    NumberViewer<uint> viewer = new NumberViewer<uint>( );
                    viewer.SetObject((LazyIO.UInt)mi);
                    panel.Controls.Add(viewer);
                    viewer.BringToFront( );
                }
                else if( mi is LazyIO.Float )
                {
                    ViewType type = ViewType.Float & viewing;
                    bool showThisType = type == ViewType.Float;
                    if( showThisType == false )
                        continue;
                    NumberViewer<float> viewer = new NumberViewer<float>( );
                    viewer.SetObject((LazyIO.Float)mi);
                    panel.Controls.Add(viewer);
                    viewer.BringToFront( );
                }
                else if( mi is StringID )
                {
                    ViewType type = ViewType.StringID & viewing;
                    bool showThisType = type == ViewType.StringID;
                    if( showThisType == false )
                        continue;
                    SIDViewer viewer = new SIDViewer(strings);
                    viewer.SetObject((StringID)mi);
                    panel.Controls.Add(viewer);
                    viewer.BringToFront( );
                }
                else if( mi is Dependency ) // before ident or else ident would pick up dependencies
                {
                    ViewType type = ViewType.Dependency & viewing;
                    bool showThisType = type == ViewType.Dependency;
                    if( showThisType == false )
                        continue;
                    IdentViewer viewer = new IdentViewer(tags);
                    viewer.SetObject((Dependency)mi);
                    panel.Controls.Add(viewer);
                    viewer.BringToFront( );
                }
                else if( mi is Ident )
                {
                    ViewType type = ViewType.Ident & viewing;
                    bool showThisType = type == ViewType.Ident;
                    if( showThisType == false )
                        continue;
                    IdentViewer viewer = new IdentViewer(tags);
                    viewer.SetObject((Ident)mi);
                    panel.Controls.Add(viewer);
                    viewer.BringToFront( );
                }
                else if( mi is BitMask8 )
                {
                    ViewType type = ViewType.BitMask8 & viewing;
                    bool showThisType = type == ViewType.BitMask8;
                    if( showThisType == false )
                        continue;
                    BitMaskViewer<byte> viewer = new BitMaskViewer<byte>( );
                    viewer.SetObject((BitMask8)mi);
                    panel.Controls.Add(viewer);
                    viewer.BringToFront( );
                }
                else if( mi is BitMask16 )
                {
                    ViewType type = ViewType.BitMask16 & viewing;
                    bool showThisType = type == ViewType.BitMask16;
                    if( showThisType == false )
                        continue;
                    BitMaskViewer<short> viewer = new BitMaskViewer<short>( );
                    viewer.SetObject((BitMask16)mi);
                    panel.Controls.Add(viewer);
                    viewer.BringToFront( );
                }
                else if( mi is BitMask32 )
                {
                    ViewType type = ViewType.BitMask32 & viewing;
                    bool showThisType = type == ViewType.BitMask32;
                    if( showThisType == false )
                        continue;
                    BitMaskViewer<int> viewer = new BitMaskViewer<int>( );
                    viewer.SetObject((BitMask32)mi);
                    panel.Controls.Add(viewer);
                    viewer.BringToFront( );
                }
                else if( mi is Enum8 )
                {
                    ViewType type = ViewType.Enum8 & viewing;
                    bool showThisType = type == ViewType.Enum8;
                    if( showThisType == false )
                        continue;
                    EnumViewer<byte> viewer = new EnumViewer<byte>( );
                    viewer.SetObject((Enum8)mi);
                    panel.Controls.Add(viewer);
                    viewer.BringToFront( );
                }
                else if( mi is Enum16 )
                {
                    ViewType type = ViewType.Enum16 & viewing;
                    bool showThisType = type == ViewType.Enum16;
                    if( showThisType == false )
                        continue;
                    EnumViewer<short> viewer = new EnumViewer<short>( );
                    viewer.SetObject((Enum16)mi);
                    panel.Controls.Add(viewer);
                    viewer.BringToFront( );
                }
                else if( mi is Enum32 )
                {
                    ViewType type = ViewType.Enum32 & viewing;
                    bool showThisType = type == ViewType.Enum32;
                    if( showThisType == false )
                        continue;
                    EnumViewer<int> viewer = new EnumViewer<int>( );
                    viewer.SetObject((Enum32)mi);
                    panel.Controls.Add(viewer);
                    viewer.BringToFront( );
                }
                else if( mi.GetType( ).IsGenericType
                    &&
                         mi.GetType( ).GetGenericTypeDefinition( ) == typeof(Reflexive<>) )
                {
                    ViewType type = ViewType.Reflexive & viewing;
                    bool showThisType = type == ViewType.Reflexive;
                    if( showThisType == false )
                        continue;
                    ReflexiveViewer viewer = new ReflexiveViewer(tags, strings);
                    viewer.SetObject((IList)mi);
                    panel.Controls.Add(viewer);
                    viewer.BringToFront( );
                }
            }
            panel.Visible = true;
            panel.EndUpdate( );
        }
        public void RepopulatePanel(HaloPanel panel)
        {
            panel.BeginUpdate( );
            foreach( Control c in panel.Controls )
            {
                Type t = c.GetType( );
                if( t.IsGenericType )
                {
                    if( t.GetGenericTypeDefinition( ) == typeof(BitMaskViewer<>) )
                    {
                        if( t.GetGenericArguments( )[0] == typeof(byte) )
                        {
                            BitMaskViewer<byte> viewer = (BitMaskViewer<byte>)c;
                            foreach( IMetaItem mi in Items )
                            {
                                if( mi.Name.Equals(viewer.ObjectName) && mi is BitMask8 )
                                {
                                    viewer.SetObject((BitMask8)mi);
                                    break;
                                }
                            }
                        }
                        else if( t.GetGenericArguments( )[0] == typeof(short) )
                        {
                            BitMaskViewer<short> viewer = (BitMaskViewer<short>)c;
                            foreach( IMetaItem mi in Items )
                            {
                                if( mi.Name.Equals(viewer.ObjectName) && mi is BitMask16 )
                                {
                                    viewer.SetObject((BitMask16)mi);
                                    break;
                                }
                            }
                        }
                        else if( t.GetGenericArguments( )[0] == typeof(int) )
                        {
                            BitMaskViewer<int> viewer = (BitMaskViewer<int>)c;
                            foreach( IMetaItem mi in Items )
                            {
                                if( mi.Name.Equals(viewer.ObjectName) && mi is BitMask32 )
                                {
                                    viewer.SetObject((BitMask32)mi);
                                    break;
                                }
                            }
                        }
                    }
                    else if( t.GetGenericTypeDefinition( ) == typeof(EnumViewer<>) )
                    {
                        if( t.GetGenericArguments( )[0] == typeof(byte) )
                        {
                            EnumViewer<byte> viewer = (EnumViewer<byte>)c;
                            foreach( IMetaItem mi in Items )
                            {
                                if( mi.Name.Equals(viewer.ObjectName) && mi is Enum8 )
                                {
                                    viewer.SetObject((Enum8)mi);
                                    break;
                                }
                            }
                        }
                        else if( t.GetGenericArguments( )[0] == typeof(short) )
                        {
                            EnumViewer<short> viewer = (EnumViewer<short>)c;
                            foreach( IMetaItem mi in Items )
                            {
                                if( mi.Name.Equals(viewer.ObjectName) && mi is Enum16 )
                                {
                                    viewer.SetObject((Enum16)mi);
                                    break;
                                }
                            }
                        }
                        else if( t.GetGenericArguments( )[0] == typeof(int) )
                        {
                            EnumViewer<int> viewer = (EnumViewer<int>)c;
                            foreach( IMetaItem mi in Items )
                            {
                                if( mi.Name.Equals(viewer.ObjectName) && mi is Enum32 )
                                {
                                    viewer.SetObject((Enum32)mi);
                                    break;
                                }
                            }
                        }
                    }
                    else if( t.GetGenericTypeDefinition( ) == typeof(NumberViewer<>) )
                    {
                        if( t.GetGenericArguments( )[0] == typeof(byte) )
                        {
                            NumberViewer<byte> viewer = (NumberViewer<byte>)c;
                            foreach( IMetaItem mi in Items )
                            {
                                if( mi.Name.Equals(viewer.ObjectName) && mi is LazyIO.Byte )
                                {
                                    viewer.SetObject((LazyIO.Byte)mi);
                                    break;
                                }
                            }
                        }
                        else if( t.GetGenericArguments( )[0] == typeof(short) )
                        {
                            NumberViewer<short> viewer = (NumberViewer<short>)c;
                            foreach( IMetaItem mi in Items )
                            {
                                if( mi.Name.Equals(viewer.ObjectName) && mi is LazyIO.Short )
                                {
                                    viewer.SetObject((LazyIO.Short)mi);
                                    break;
                                }
                            }
                        }
                        else if( t.GetGenericArguments( )[0] == typeof(ushort) )
                        {
                            NumberViewer<ushort> viewer = (NumberViewer<ushort>)c;
                            foreach( IMetaItem mi in Items )
                            {
                                if( mi.Name.Equals(viewer.ObjectName) && mi is LazyIO.UShort )
                                {
                                    viewer.SetObject((LazyIO.UShort)mi);
                                    break;
                                }
                            }
                        }
                        else if( t.GetGenericArguments( )[0] == typeof(int) )
                        {
                            NumberViewer<int> viewer = (NumberViewer<int>)c;
                            foreach( IMetaItem mi in Items )
                            {
                                if( mi.Name.Equals(viewer.ObjectName) && mi is LazyIO.Int )
                                {
                                    viewer.SetObject((LazyIO.Int)mi);
                                    break;
                                }
                            }
                        }
                        else if( t.GetGenericArguments( )[0] == typeof(uint) )
                        {
                            NumberViewer<uint> viewer = (NumberViewer<uint>)c;
                            foreach( IMetaItem mi in Items )
                            {
                                if( mi.Name.Equals(viewer.ObjectName) && mi is LazyIO.UInt )
                                {
                                    viewer.SetObject((LazyIO.UInt)mi);
                                    break;
                                }
                            }
                        }
                        else if( t.GetGenericArguments( )[0] == typeof(float) )
                        {
                            NumberViewer<float> viewer = (NumberViewer<float>)c;
                            foreach( IMetaItem mi in Items )
                            {
                                if( mi.Name.Equals(viewer.ObjectName) && mi is LazyIO.Float )
                                {
                                    viewer.SetObject((LazyIO.Float)mi);
                                    break;
                                }
                            }
                        }
                    }
                }
                if( c is IdentViewer )
                {
                    IdentViewer viewer = (IdentViewer)c;
                    foreach( IMetaItem mi in Items )
                    {
                        if( mi.Name.Equals(viewer.ObjectName) && mi is Ident ) // covers reflexives too
                        {
                            viewer.SetObject((Ident)mi);
                            break;
                        }
                    }
                }
                else if( c is SIDViewer )
                {
                    SIDViewer viewer = (SIDViewer)c;
                    foreach( IMetaItem mi in Items )
                    {
                        if( mi.Name.Equals(viewer.ObjectName) && mi is StringID )
                        {
                            viewer.SetObject((StringID)mi);
                            break;
                        }
                    }
                }
                else if( c is ReflexiveViewer )
                {
                    ReflexiveViewer viewer = (ReflexiveViewer)c;
                    foreach( IMetaItem mi in Items )
                    {
                        if( mi.Name.Equals(viewer.ObjectName)
                            && mi.GetType( ).IsGenericType &&
                            mi.GetType( ).GetGenericTypeDefinition( ) == typeof(Reflexive<>) )
                        {
                            viewer.SetObject((IList)mi);
                            break;
                        }
                    }
                }
            }
            panel.EndUpdate( );
        }

        #endregion
    }
}
