﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Utilities;

namespace HaloMap
{
    public class TagList : List<Tag>
    {
        #region properties

        /// <summary>
        /// Gets the offset of the FileTable
        /// </summary>
        public int Offset
        {
            get { return Header.FileTableOffset; }
        }
        /// <summary>
        /// Gets the offset of the Index
        /// </summary>
        public int IndexOffset
        {
            get { return Header.FilesIndexOffset; }
        }
        /// <summary>
        /// Gets the number of Tags in the FileTable
        /// </summary>
        public int TagCount
        {
            get { return this.Count; }
        }
        /// <summary>
        /// Gets the offset of the Object Index
        /// </summary>
        private int ObjectIndexOffset
        {
            get { return Index.ObjectIndexOffset; }
        }
        /// <summary>
        /// Gets the size of the FileTable
        /// </summary>
        private int Size
        {
            get { return Header.FileTableSize; }
        }

        private Dictionary<int, Tag> TagByIdent { get; set; }
        private Dictionary<string, Tag> TagByPath { get; set; }
        private BiDirectionalHashtable<Tag, TreeNode> TagsAndTreeNodes { get; set; }

        private MapHeader Header
        {
            get;
            set;
        }
        private MapIndex Index
        {
            get;
            set;
        }

        #endregion
        #region constructor

        public TagList(MapHeader mh, MapIndex mi)
            : base(6000) // for good measure(little resizing)
        {
            Header = mh;
            Index = mi;

            TagTypes = new List<string>();
            TagByIdent = new Dictionary<int, Tag>();
            TagByPath = new Dictionary<string, Tag>();
            TagsAndTreeNodes = new BiDirectionalHashtable<Tag, TreeNode>();
        }

        #endregion
        #region methods

        public void Read(BinaryReader br)
        {
            br.BaseStream.Position = ObjectIndexOffset;
            int count = Index.ObjectCount;
            string[] tagType = new string[count];
            int[] ident = new int[count];
            int[] rawMetaOffset = new int[count];
            int[] metaSize = new int[count];

            for (int i = 0; i < count; i++)
            {
                char[] temp = br.ReadChars(4);
                Array.Reverse(temp);
                tagType[i] = new string(temp);
                // if it's not in the list of tagClass types then add it
                if (TagTypes.Contains(tagType[i]) == false)
                {
                    TagTypes.Add(tagType[i]);
                }
                ident[i] = br.ReadInt32();
                rawMetaOffset[i] = br.ReadInt32();
                metaSize[i] = br.ReadInt32();
            }
            TagTypes.Sort();

            string[] tagName = new string[count];

            br.BaseStream.Position = Offset;

            char c;// temp char
            for (int i = 0; i < count; i++)
            {
                StringBuilder sb = new StringBuilder();
                while ((c = br.ReadChar()) != '\0')
                {
                    sb.Append(c);
                }
                tagName[i] = sb.ToString();
                Tag t = new Tag(ident[i], rawMetaOffset[i], metaSize[i],
                    tagType[i].Equals("sbsp") || tagType[i].Equals("ltmp") ?
                    Index.PrimaryMagic : Index.SecondaryMagic, tagType[i], tagName[i]);
                TagByIdent.Add(t.Ident, t);
                string key = string.Format("{0}.{1}", t.Name, t.Type);
                TagByPath.Add(key, t);
                Add(t);
                
           }
        }
        public List<string> TagTypes { get; protected set; }
        public string[] GetAllTagsOfType(string tagType)
        {
            if (TagTypes.Contains(tagType) == false) return null;
            List<string> names = new List<string>();
            foreach (Tag t in this)
            {
                if (t.Type.Equals(tagType))
                {
                    names.Add(t.Name);
                }
            }
            return names.ToArray();
        }
        /// <summary>
        /// Finds the TagClass by the offset in the map
        /// </summary>
        /// <param name="offset">The map offset</param>
        /// <returns>The tagClass or null if no tagClass is at the offset</returns>
        public Tag FindTagByOffset(int offset)
        {
            for (int i = 0; i < Count; i++)
            {
                Tag t = this[i];
                if (offset >= t.MetaOffset && offset < (t.MetaOffset + t.MetaSize))
                {
                    return t;
                }
            }
            // you made it through them all and it wasn't in the list
            return null;
        }
        /// <summary>
        /// Finds the TagClass by the Ident
        /// </summary>
        /// <param name="ident">The ident of the TagClass to find</param>
        /// <returns>The TagClass with the specified ident or null</returns>
        public Tag FindTag(int ident)
        {
            Tag temp;
            if (TagByIdent.TryGetValue(ident, out temp))
            {
                return temp;
            }
            return null;
        }
        /// <summary>
        /// Finds the TagClass with the specified tagClass path
        /// </summary>
        /// <param name="type">The tagClass type</param>
        /// <param name="name">The tagClass name</param>
        /// <returns>The TagClass with the specified tagClass path or null</returns>
        public Tag FindTag(string type, string name)
        {
            //string key = string.Format("{0}.{1}", name, type);
            //Tag temp;
            //if (TagByPath.TryGetValue(key, out temp))
            //{
            //    return temp;
            //}
            foreach (Tag t in this)
            {
                if (t.Type.Equals(type)
                    &&
                    t.Name.Equals(name))
                    return t;
            }
            return null;
        }
        /// <summary>
        /// Finds the TagClass associated with the TreeNode
        /// </summary>
        /// <param name="tn">The TreeNode to find the TagClass for</param>
        /// <returns>The TagClass associated with the TreeNode or null if none exists</returns>
        public Tag FindTag(TreeNode tn)
        {
            return TagsAndTreeNodes.GetKey(tn);
        }
        public void LoadTreeView(TreeView tv)
        {
            tv.BeginUpdate();
            tv.Nodes.Clear();
            TreeNode top = new TreeNode(Header.InternalName);
            TreeNode[] nodes = LoadTreeView();
            top.Nodes.AddRange(nodes);
            tv.Nodes.Add(top);
            //tv.Nodes.Add(new TreeNode(Header.InternalName, LoadTreeView()));
            tv.Sort();
            tv.Nodes[0].Expand();
            tv.EndUpdate();
        }
        /// <summary>
        /// Loads the tags into TreeNodes in order to populate the TreeView
        /// </summary>
        /// <returns>A TreeNode Array containing all of the tagClass types and tags of the map</returns>
        TreeNode[] LoadTreeView()
        {
            TagsAndTreeNodes = new BiDirectionalHashtable<Tag, TreeNode>();
            TreeNode[] tagTypeNodes = new TreeNode[TagTypes.Count];
            for (int x = 0; x < TagTypes.Count; x++)
            {
                tagTypeNodes[x] = new TreeNode(TagTypes[x], 1, 1);
                for (int y = 0; y < this.Count; y++)
                {
                    Tag t = this[y];
                    if (t.Type.Equals(TagTypes[x]))
                    {
                        TreeNode tn = new TreeNode(t.Name, 2, 2);
                        tagTypeNodes[x].Nodes.Add(tn);
                        TagsAndTreeNodes.AddItem(t, tn);
                    }
                }
            }
            return tagTypeNodes;
        }
        /// <summary>
        /// Finds the TreeNode associated with the specified TagClass
        /// </summary>
        /// <param name="tagClass">The TagClass to find the TreeNode for</param>
        /// <returns>The TreeNode associated with the specified TagClass or null</returns>
        public TreeNode FindTreeNode(Tag tag)
        {
            return TagsAndTreeNodes.GetItem(tag);
        }

        #endregion
    }
}
